from pyepl.locals import *             # Import PyEPL
from copy import *
import random
import math

def makeLayout(grid, storepool, buildingpool):
    """
    """
    # start with an empty list for the layout configuration
    config = []

    # ...
    for row in grid:
        for spot in row:
            if spot not in ("S", "B", "N", "NBNSW", "E"):
                try:
                    spot = storepool.findBy(name = spot).copy()
                    spot.type = "store"
                except LookupError:
                    try:
                        spot = buildingpool.findBy(name = spot).copy()
                        spot.type = "building"
                    except LookupError:
                        raise LookupError, "No store or building available named %s" % spot
            config.append(spot)
    layout = YCLayout(len(grid[0]), len(grid), config)
    layout.fillBlanks(storepool.sample(), buildingpool.sample())
    return layout

def chooseStores(config, environmentPool):
    """
    """
    # if useStores is in the config...
    if hasattr(config, "useStores"):
        avail_stores = Pool()
        for name in config.useStores:
            avail_stores.append(config.stores.findBy(name = name))
    else:
        avail_stores = config.stores
    
    # if useLayout is in the config...
    if hasattr(config, "useLayout"):
        # make a layout using the specified values
        layout = makeLayout(config.useLayout, avail_stores, config.buildings)

        # return the stores selected
        return layout.getStorePool()
    
    # if useStoresFrom is in the config...
    if hasattr(config, "useStoresFrom"):
        # return a copy of the store pool from the indicated environment
        return environmentPool.findBy(name = config.useStoresFrom).stores[:]
    
    # make a copy of the available stores
    storepool = avail_stores[:]

    # if excludeStoresFrom is in the config...
    if hasattr(config, "excludeStoresFrom"):
        # start a Pool of items to be exluded
        exclude = Pool()

        # populate the exclude pool...
        for cityname in config.excludeStoresFrom:
            exclude.extend(environmentPool.findBy(name = cityname).stores)

        # find where items in the exclude Pool match items in the store pool...
        for x in exclude:
            # ...and remove them
            try:
                storepool.remove(x)
            except ValueError:
                # if it's not already there, fine
                pass

    # return a sample of the stores now in the store pool
    return storepool.sample(config.numStores)

def chooseBuildings(config, environmentPool):
    """
    """
    # if useBuildings is in the config...
    if hasattr(config, "useBuildings"):
        avail_buildings = Pool()
        for name in config.useBuildings:
            avail_buildings.append(config.buildings.findBy(name = name))
    else:
        avail_buildings = config.buildings
    
    # if useLayout is in the config...
    if hasattr(config, "useLayout"):
        # make a layout using the specified values
        layout = makeLayout(config.useLayout, config.stores, avail_buildings)

        # return the buildings selected
        return layout.getBuildingPool()
    
    # if we're using null buildings...
    if config.nullBuildings:
        # ... then just return an empty pool
        return Pool()

    # if useBuildingsFrom is in the config...
    if hasattr(config, "useBuildingsFrom"):
        # return a copy of the building pool from the indicated environment
        return environmentPool.findBy(name = config.useBuildingsFrom).buildings[:]
    
    # make a copy of the global building pool
    buildingpool = avail_buildings[:]

    # if excludeBuildingsFrom is in the config...
    if hasattr(config, "excludeBuildingsFrom"):
        # start a Pool of items to be exluded
        exclude = Pool()

        # populate the exclude pool...
        for cityname in config.excludeBuildingsFrom:
            exclude.extend(environmentPool.findBy(name = cityname).buildings)

        # find where items in the exclude Pool match items in the building pool...
        for x in exclude:
            # ...and remove them
            try:
                buildingpool.remove(x)
            except IndexError:
                # if it's not already there, fine
                pass

    # return a sample of the buildings now in the building pool
    return buildingpool.sample((config.cityBlocks[0] * config.cityBlocks[1]) - config.numStores)

def choosePassengers(config, environmentPool):
    """
    """
    # start with an empty list for deliveries
    deliveries = []
    
    # for each delivery...
    for deliveryNum in xrange(config.numDeliveries):
        # get delivery-specific configuration
        deliveryconfig = config.sequence(deliveryNum)

        subpool = Pool()
        for n in xrange(deliveryconfig.passengersPerDelivery):
            if hasattr(deliveryconfig, "useDelivery"):
                pname = deliveryconfig.useDelivery[n][0]
            else:
                pname = None

            # if we have a predetermined passenger...
            if pname:
                subpool.append(deliveryconfig.passengerIdentities.findBy(name = pname))

            # use passengers from another environment, if requested...
            elif hasattr(deliveryconfig, "usePassengersFrom"):
                subpool.append(environmentPool.findBy(name = config.usePassengersFrom).passengers[deliveryNum][n])

            # if we want to use the same passengers for each delivery and this isn't the first...
            elif deliveryconfig.samePassengersEachDelivery and deliveryNum:
                subpool.append(deliveries.append(deliveries[-1][n]))

            # otherwise, use random passengers...
            else:
                # limit passengers to choose from?
                if hasattr(deliveryconfig, "usePassengers"):
                    passengerpool = Pool()
                    for pname in deliveryconfig.usePassengers:
                        passengerpool.append(deliveryconfig.passengerIdentities.findBy(name = pname))
                else:
                    # make a copy of the passenger pool
                    passengerpool = deliveryconfig.passengerIdentities[:]

                # start a Pool of items to be exluded, populated with passengers already placed
                exclude = subpool[:]

                # if excludePassengersFrom is in the config...
                if hasattr(config, "excludePassengersFrom"):
                    # populate the exclude pool...
                    for cityname in config.excludePassengersFrom:
                        exclude.extend(environmentPool.findBy(name = config.usePassengersFrom).passengers[deliveryNum])

                # find where items in the exclude Pool match items in the passenger pool...
                for x in exclude:
                    # ...and remove them
                    try:
                        passengerpool.remove(x)
                    except IndexError:
                        # if it's not already there, fine
                        pass

                # select a passenger and append it to the subpool
                subpool.append(passengerpool.randomChoice())

        # append the selected passengers as a delivery
        deliveries.append(subpool)

    # return the delivery list
    return deliveries

def nextLayout(config, history, stores, buildings):
    """
    """
    # if useLayout is in the config...
    if hasattr(config, "useLayout"):
        # make a layout using the specified values
        layout = makeLayout(config.useLayout, stores, buildings)

        # add physical measurements to the layout
        layout.addMeasurements(config.unitScale, config.wallHeight, config.curbHeight)
        
        # return the layout
        return layout
    
    # unpack the city dimensions
    xsize, ysize = config.cityBlocks
    
    # useCityGeometryFrom?
    if hasattr(config, "useCityGeometryFrom"):
        # get the geometry to use
        geom = history.findCity(config.useCityGeometryFrom).getLayout().getGeometry()

        # create a layout of it
        layout = YCLayout(xsize, ysize, geom)

        # otherwise, create a new city geometry...
    else:
	#generate a list of constraint sizes
	def getSizes(blocks, numStores):
		'''
		Generate a list of lists.  Each element of each list corresponds to
		the area within the environment to which the corresponding store
		should be constrained.  Initially, each store is constrained only by
		the boundaries of the environment.
		'''
		#define some helper methods
		def inc(v):
			'''
			Increment a vector, elementwise, counting in 2-space with a
			blocks-ary number system.  Truncate at len(v).  Also, this number
			system's lowest placeholder digit is on the left.
			'''
			if len(v) == 0:
				return v
			else:
				x = v[0][0]
				y = v[0][1]
				
				if not (x == blocks[0] and y == blocks[1]):
					if not x == blocks[0]:
						x = x + 1
					if not y == blocks[1]:
						y = y + 1
					return [(x,y)]+v[1:]
				else:
					return [(0,0)]+inc(v[1:])
		def inc2(v):
			'''
			This method behaves similarly to inc(v), but once a "digit" reaches
			the maximum value (blocks), the value for that digit sticks.
			'''
			if len(v) == 0:
				return v
			else:
				x = v[0][0]
				y = v[0][1]
				
				if not (x == blocks[0] or y == blocks[1]):
					if not x == blocks[0]:
						x = x + 1
					if not y == blocks[1]:
						y = y + 1
					return [(x,y)]+v[1:]
				else:
					return [blocks]+inc2(v[1:])
				
		def sorted(myList):
			'''
			Is myList sorted?
			'''
			c = copy(myList)
			c.sort()
			if c == myList:
				return True
			else:
				return False
				
		def getMinDistance():
			'''
			Return the minimum distance which adjacent "constraint rings"
			must be from one another, given the current environment size.
			'''
			if(blocks > (4,4)):
				return 3
			else:
				return 2
				
		##########################
		# make the list of sizes #
		##########################
		sizes = []
		this = []
		currentMin = (0,0) #current store constraint
		dist = getMinDistance() #distance to separate constrained sectors
		
		#compute the most constrained vector possible, given the environment
		#size
		for i in range(numStores):
			this.append(currentMin)
			for j in range(dist):
				currentMin = inc2([currentMin])[0]
				
		#comute the least constrained vector possible, given the environment
		#size
		maxVector = []
		for i in range(numStores):
			maxVector.append(blocks)
			
		#create the list of sizes (generated in reverse order)
		while not this == maxVector:
			#ensure that store n is constrained as much or more than
			#store (n+1).
			if(sorted(this)):
				sizes.append(this)
			this = copy(this)
			this = inc(this)
		
		sizes.append(maxVector)
		
		sizes.reverse()
		
		return sizes    
			
	#generate a group of stores
	def getStores(blocks,numStores):
		'''
		Using the list of lists generated by the getSizes method, try to place
		numStores stores within an environment of size blocks.  Return a list of
		length numStores containing the coordinates of stores which were 
		successfully placed.  If the stores cannot be placed within the environment,
		this method throws an error.
		'''
		
		def getCoords(dims):
			'''
			Generate a list of coordinates, given an ordered pair of
			dimensions to use
			'''
			coords = []
			for y in range(dims[1]):
				for x in range(dims[0]):
					coords.append((x,y))
			return coords
		
		def getAllCoords(dimList):
			'''
			Given an array of sizes, return an array of lists of
			all positions constrained within the corresponding size
			'''
			allCoords = []
			for i in range(len(dimList)):
				allCoords.append(getCoords(dimList[i]))
			return allCoords
		
		def pruneCoords(available,stores):
			'''
			Given the current constraints (flags) on store placement, and
			the list of stores placed so far, remove positions at which
			future stores may not be located.
			'''
			
			#flags
			cornersFlag = config.noStoresInAllCorners
			neighborsFlag = config.noStoresNearestNeighbors
			rowsAndColumnsFlag = config.noStoreFilledRowsOrColumns
			
			#the most recently placed store
			lastStore = stores[-1]
			x = lastStore[0]
			y = lastStore[1]
			
			#keep a list of positions to remove (initially contains
			#the last store to be placed)
			removeCoords = [lastStore]
			
			if(cornersFlag and len(stores) >= 3):
				#list the environment's corners
				corners = [(0,0), (blocks[0]-1,0), (0,blocks[1]-1), (blocks[0]-1,blocks[1]-1)]
				
				#which corners have not been used yet?
				unUsedCorners = []
				for c in corners:
					if c not in stores:
						unUsedCorners.append(c)
				
				#if only one corner is left, add it to the list of coordinates
				#to remove
				if len(unUsedCorners) == 1:
					removeCoords.append(unUsedCorners[0])
			
			if(neighborsFlag):
				if(blocks < (4,4)):
					pass
				else:
					for c in available:
						#remove adjacent blocks
						if((abs(x - c[0]) < 2) and (abs(y - c[1]) < 2)):
							removeCoords.append(c)
						#ensure 2 blocks horizontally and vertically
						if(blocks >= (5,5) and not blocks >= (7,7)):
							if(((abs(x - c[0]) < 3) and (y == c[1])) or \
							   ((abs(y - c[1]) < 3) and (x == c[0]))):
								removeCoords.append(c)
						#ensure knights moves are empty as well as the above
						if(blocks >= (7,7)):
							if((abs(x - c[0]) < 3) and (abs(y - c[1]) < 3)):
								removeCoords.append(c)
			
			if(rowsAndColumnsFlag and len(stores) == numStores - 1):
				#append unique xs and ys to a list.  If either list contains
				#only one element, remove all other coordinates with the same
				#x or y, respectively.
				xs = []
				ys = []
				for s in stores:
					if s[0] not in xs:
						xs.append(s[0])
					if s[1] not in ys:
						ys.append(s[1])
				if len(xs) == 1:
					#remove all coordinates with the same x value
					for c in available:
						if c[0] == xs[0]:
							removeCoords.append(c)
				
				if len(ys) == 1:
					#remove all coordinates with the same y value
					for c in available:
						if c[1] == ys[0]:
							removeCoords.append(c)
				
			#remove all coordinates in removeCoords from "available" list
			for c in removeCoords:
				try:
					available.remove(c)
				except ValueError:
					pass
					
			#return the new (updated) list of available coordinates
			return available
		
		def pruneAllCoords(coordsLists, stores):
			'''
			For each list of coordinates, remove coordinates as appropriate
			according to stores placed so far and constraints (flags) listed
			in the configuration file.
			'''
			for cl in coordsLists:
				cl = pruneCoords(cl, stores)
			
			return coordsLists
		
		def printEnvironment(stores):
			'''
			Use for debugging; prints a text-based representation of
			the environment generated, where stores are marked by X's.
			Buildings are not shown.  This will be messy if the terminal
			window's width is smaller than (x*3)+2 characters, where x 
			is the environment's x-dimension, or if a non-uniformly-spaced
			font is used.		
			'''
			x = blocks[0]
			y = blocks[1]
			
			#print a summary:
			print
			print 'SIZE: '+str(blocks)
			print 'STORES: '+str(stores)
			print
			
			#print the top row of the diagram
			print '  |',
			for i in range(x):
				print i,
				print '|',
					
			#print the columns (one row at a time)
			for i in range(y):
				#start a new line
				print
				#print a divider
				for j in range((x*2)+2):
					print '-',
				#start a new line
				print
				#print row number
				print i,
				#print the row
				print '|',
				for j in range(x):
					if (j,i) in stores:
						print 'X |',
					else:
						print '  |',
			#start a new line
			print
			#print the last divider:
			for i in range((x*2)+2):
				print '-',
			#start a new line, and skip another
			print
			print
			
		##########################################
		# attempt to generate the list of stores #
		##########################################
		for s in getSizes(blocks,numStores):
			#generate a list of coordinates to which each store
			#is currently constrained
			allCoords = getAllCoords(s)
			
			#place one store at a time, and store them
			placedStores = []
			
			while(len(placedStores) < numStores):
				#keep track of the current store's number, for convenience
				myNum = len(placedStores)
				
				#attempt to place the current store randomly within
				#its constrained area.
				if(len(allCoords[myNum]) > 0):
					placedStores.append(random.choice(allCoords[myNum]))
					#have all stores been placed now?
					if(len(placedStores) == numStores):
						#print out a pretty picture!
						printEnvironment(placedStores)
						return placedStores
					else:
						#prune allCoords as needed
						allCoords = pruneAllCoords(allCoords,placedStores)
				else:
					break
		#If the for loop finishes without placing all stores, throw an
		#error: we were unable to place the stores.  NOTE: this does NOT mean
		#that placing stores could NEVER be done correctly; the algorithm
		#places stores according to a "best guess," but is not guaranteed to
		#find a solution (a) as quickly as possible in all cases, or (b) in
		#all cases.  When the number of stores is less than min(m,n) for an m
		#by n environment, the algorithm should place stores correctly.
		#   In fact, running the algorithm again for the same parameters may result
		#in success!
		raise "\n\nCould not place all stores given current config file."
        
        placedStores = getStores(config.cityBlocks,config.numStores)
	
	#now stores are all placed...
	#generate a "layout string" to pass to the YCLayout constructor
	#re-create the list of possible coordinates

	#we need to generate coordinates in this order for when we create
	#the geometry string
	coords = []
	for y in range(config.cityBlocks[1]):
		for x in range(config.cityBlocks[0]):
			coords.append((x,y))

        geom = []
	for c in coords:
            if c in placedStores:
                geom.append('S')
            elif config.nullBuildings:
                geom.append('N')
            else:
                geom.append('B')
 
	#create the city layout
	layout = YCLayout(xsize,ysize,geom)

    # useStoreGeometryFrom?
    if hasattr(config, "useStoreGeometryFrom"):
        # get the store geometry to use
        storegeom = history.findCity(config.useCityGeometryFrom).getLayout().getStoreGeometry()

        # rearrange the order of the stores list to match the order of storegeom...
        for store in storegeom:
            try:
                # find the store in the stores list
                n = stores.index(store)

                # remove the store
                s = stores.pop(n)

                # and put it at the end
                stores.append(s)
            except ValueError:
                # if the store is not in the stores list, go on
                # such stores will be groups together in their original relative order at the beginning of the list
                pass

    # useBuildingGeometryFrom?
    if hasattr(config, "useBuildingGeometryFrom"):
        # get the building geometry to use
        buildinggeom = history.findCity(config.useCityGeometryFrom).getLayout().getBuildingGeometry()

        # rearrange the order of the buildings list to match the order of buildinggeom...
        for building in buildinggeom:
            try:
                # find the building in the buildings list
                n = buildings.index(building)

                # remove the building
                s = buildings.pop(n)

                # and put it at the end
                buildings.append(s)
            except ValueError:
                # if the building is not in the buildings list, go on
                # such buildings will be groups together in their original relative order at the beginning of the list
                pass

    # put actual stores and buildings into the layout
    layout.fillBlanks(stores, buildings)

    # add physical measurements to the layout
    layout.addMeasurements(config.unitScale, config.wallHeight, config.curbHeight)

    # return the layout
    return layout



def nextPassengersWaiting(config, history, identities, deliveryNum, avatarPosition):
    """
    """
    # check the layout  :-)
    layout = history[-1].layout
    
    # start with an empty list for placed passengers
    placed = []

    # start with an empty Pool for passenger placement boxes
    possible_boxes = Pool()
    
    # create perimeter boxes along which passengers could possibly be placed...
    cityrows = layout.getHeight()
    citycolumns = layout.getWidth()
    unitscale = layout.unitscale
    lowx = (cityrows - 1) * 3 * unitscale / 2
    lowz = (citycolumns - 1) * 3 * unitscale / 2
    possible_destinations = Pool()
    for rownum in xrange(cityrows):
        for columnnum in xrange(citycolumns):
            item = layout[columnnum, rownum]
            xpos = (unitscale * 3 * rownum) - lowx
            zpos = (unitscale * 3 * columnnum) - lowz
            if item.type == "store":
                possible_destinations.append(name = item.name,
                                             item = item,
                                             rownum = rownum,
                                             columnnum = columnnum
                                             )
                if config.passengersWaitStoreBoxSize:
                    halfbox = config.passengersWaitStoreBoxSize * unitscale / 2
                    store_pass_alt = config.passengersWaitStoreAltitude
                    box = PoolDict(
                        name = "%d_%d" % (rownum, columnnum),
                        north = RPLineSegment(1.0, (xpos - halfbox, store_pass_alt, zpos - halfbox), (xpos + halfbox, store_pass_alt, zpos - halfbox)),
                        east = RPLineSegment(1.0, (xpos + halfbox, store_pass_alt, zpos - halfbox), (xpos + halfbox, store_pass_alt, zpos + halfbox)),
                        south = RPLineSegment(1.0, (xpos + halfbox, store_pass_alt, zpos + halfbox), (xpos - halfbox, store_pass_alt, zpos + halfbox)),
                        west = RPLineSegment(1.0, (xpos - halfbox, store_pass_alt, zpos + halfbox), (xpos - halfbox, store_pass_alt, zpos - halfbox)),
                        item = item,
                        rownum = rownum,
                        columnnum = columnnum
                        )
                    possible_boxes.append(box)
            elif item.type == "building" or item.type == "nullbuilding":
                if config.passengersWaitBuildingBoxSize:
                    halfbox = config.passengersWaitBuildingBoxSize * unitscale / 2
                    building_pass_alt = config.passengersWaitBuildingAltitude
                    box = PoolDict(
                        name = "%d_%d" % (rownum, columnnum),
                        north = RPLineSegment(1.0, (xpos - halfbox, building_pass_alt, zpos - halfbox), (xpos + halfbox, building_pass_alt, zpos - halfbox)),
                        east = RPLineSegment(1.0, (xpos + halfbox, building_pass_alt, zpos - halfbox), (xpos + halfbox, building_pass_alt, zpos + halfbox)),
                        south = RPLineSegment(1.0, (xpos + halfbox, building_pass_alt, zpos + halfbox), (xpos - halfbox, building_pass_alt, zpos + halfbox)),
                        west = RPLineSegment(1.0, (xpos - halfbox, building_pass_alt, zpos + halfbox), (xpos - halfbox, building_pass_alt, zpos - halfbox)),
                        item = item,
                        rownum = rownum,
                        columnnum = columnnum
                        )
                    possible_boxes.append(box)
# with no sidewalks passengers still wait around where buildings would be, but need to have their feet on the ground
            elif item.type == "nullbuildingnullsidewalk":
                if config.passengersWaitBuildingBoxSize:
                    halfbox = config.passengersWaitBuildingBoxSize * unitscale / 2
                    building_pass_alt = 0.0
                    box = PoolDict(
                        name = "%d_%d" % (rownum, columnnum),
                        north = RPLineSegment(1.0, (xpos - halfbox, building_pass_alt, zpos - halfbox), (xpos + halfbox, building_pass_alt, zpos - halfbox)),
                        east = RPLineSegment(1.0, (xpos + halfbox, building_pass_alt, zpos - halfbox), (xpos + halfbox, building_pass_alt, zpos + halfbox)),
                        south = RPLineSegment(1.0, (xpos + halfbox, building_pass_alt, zpos + halfbox), (xpos - halfbox, building_pass_alt, zpos + halfbox)),
                        west = RPLineSegment(1.0, (xpos - halfbox, building_pass_alt, zpos + halfbox), (xpos - halfbox, building_pass_alt, zpos - halfbox)),
                        item = item,
                        rownum = rownum,
                        columnnum = columnnum
                        )
                    possible_boxes.append(box)

    # eliminate boxes, or parts of boxes, that would violate constraints given the delivery history and current position...
    #...

    # restrict possible destinations for all passengers in this delivery...
    # do we use the allStoresBeforeRepeat constraint?
    if config.allStoresBeforeRepeat and not (hasattr(config, "useDelivery") and config.useDelivery[n][2]):
        # if so, use a modulus function to decide which step of the cycle we're on...
        cyclepos = deliveryNum % config.numStores

        # we only need to eliminate stores if we're not on step zero...
        if cyclepos:
            # slice the history to get the previous deliveries within this cycle
            recent = history[-1][-cyclepos:]

            # iterate through those...
            for used in recent:
                # removing the used stores from the current possible destinations list
                possible_destinations.removeBy(item = used.getDestination())

    # onePassengerOutInFront?
    if config.onePassengerOutInFront:
        # select a random position in the identities list to be the "one passenger"
        one_passenger = random.randrange(len(identities))
    else:
        # otherwise, there is no "one passenger"
        one_passenger = None

    # place each passenger identity, eliminating boxes or parts of boxes as constraints require...
    for n, i in enumerate(identities):
        # use predetermined destination?
        if hasattr(config, "useDelivery") and config.useDelivery[n][2]:
            destination = possible_destinations.findBy(name = config.useDelivery[n][2])
        else:
            # copy possible destinations for use only with this passenger
            destinations = possible_destinations[:]
            
            # limit possible destinations based on constraints FOR THIS PASSENGER ONLY...
            #...
            
            # pick a destination
            destination = destinations.randomChoice()

        # is this the "one passenger"?...
        if n == one_passenger:
            # He is the one.  Put him at the correct distance directly in front of the avatar...
            yawrad = math.radians(avatarPosition[3])
            position = (avatarPosition[0] + (math.sin(yawrad) * config.onePassengerOutInFront),
                        0.0,
                        avatarPosition[2] - (math.cos(yawrad) * config.onePassengerOutInFront)
                        )

        # is it a predetermined position?...
        elif hasattr(config, "useDelivery") and config.useDelivery[n][1]:
            position = config.useDelivery[n][1]
        else:
            # otherwise, select position normally...

            # copy the box array for use only with this passenger
            boxes = possible_boxes.middleDepthCopy()

            # limit possible passenger positions FOR THIS PASSENGER ONLY based on destination constraints...
            # passengerObscureDestination?
            if config.passengerObscureDestination:
                for box in boxes:
                    if box.rownum <= destination.rownum:
                        box.east = RPSpace()
                    if box.rownum >= destination.rownum:
                        box.west = RPSpace()
                    if box.columnnum <= destination.columnnum:
                        box.south = RPSpace()
                    if box.columnnum >= destination.columnnum:
                        box.north = RPSpace()

            # pick a position...
            pickup_space = RPMutableSpace()
            for box in boxes:
                pickup_space.add(box.north)
                pickup_space.add(box.east)
                pickup_space.add(box.south)
                pickup_space.add(box.west)
            position, source = pickup_space.generatePositionAndSource()

            # limit positions for remaining passengers based on selection...
            # onePassengerPerBlock constraint?
            if config.onePassengerPerBlock:
                # look at all the boxes
                for box in possible_boxes:
                    # was the passenger placed on this box?
                    if source in (box.north, box.east, box.south, box.west):
                        # if so, replace the sides of the box with zero-weighted RPSpaces so they won't be used any more...
                        box.north = RPSpace()
                        box.east = RPSpace()
                        box.south = RPSpace()
                        box.west = RPSpace()

                        # the passenger was only placed on one box, so we can stop looking
                        break

        # add identity, position, and destination for the placed list
        placed.append((i, position, destination.item))

    # return the list of placed passengers
    return placed

def nextStartingPosition(config, layout, history):
    """
    """
    if hasattr(config, "useStartingPosition"):
        return config.useStartingPosition
    avatar_space = RPMutableSpace()
    xmin = layout.realsize[0] / -2
    ymin = layout.realsize[1] / -2
    xroads = []
    yroads = []
    for x in xrange(layout.getWidth() + 1):
        xp = (x * 3 * config.unitScale) + xmin + (config.unitScale / 2)
        xroads.append(RPLineSegment(1.0, (xp, 0.0, ymin + config.avatarRadius), (xp, 0.0, ymin + layout.realsize[1] - config.avatarRadius)))
    for y in xrange(layout.getHeight() + 1):
        yp = (y * 3 * config.unitScale) + ymin + (config.unitScale / 2)
        yroads.append(RPLineSegment(1.0, (xmin + config.avatarRadius, 0.0, yp), (xmin + layout.realsize[0] - config.avatarRadius, 0.0, yp)))
    avatar_space.add(*xroads)
    avatar_space.add(*yroads)
    position, source = avatar_space.generatePositionAndSource()
    if not config.startOrientedAlongRoad:
        yaw = random.uniform(0.0, 360.0)
    elif source in xroads:
        if abs(position[2] - ymin) < config.minimumDistanceFromFacedWall:
            yaw = 180.0
        elif abs((ymin + layout.realsize[1]) - position[2]) < config.minimumDistanceFromFacedWall:
            yaw = 0.0
        else:
            yaw = random.choice((0.0, 180.0))
    elif source in yroads:
        if abs(position[0] - xmin) < config.minimumDistanceFromFacedWall:
            yaw = 90.0
        elif abs((xmin + layout.realsize[0]) - position[0]) < config.minimumDistanceFromFacedWall:
            yaw = 270.0
        else:
            yaw = random.choice((90.0, 270.0))
    return (position[0], position[1], position[2], yaw, 0.0, 0.0)

class YCHistory:
    """
    """
    def __init__(self):
        """
        """
        # start with an empty list for cities
        self.cities = []
    def __len__(self):
        """
        """
        return len(self.cities)
    def __getitem__(self, index):
        """
        """
        # just index the cities list
        return self.cities[index]
    def city(self, name, layout, startingposition):
        """
        """
        # create a city record from the provided arguments
        c = YCHistoryCity(name, layout, startingposition)

        # append it to the cities list
        self.cities.append(c)

        # return it
        return c
    def findCity(self, name):
        """
        """
        # search through the cities for one with the given name...
        for c in self.cities:
            # ...when we find it...
            if c.getName() == name:
                # ...return it
                return c

        # if we didn't find it, raise an exception
        raise LookupError, "No city by name %s" % name
    def delivery(self, idents):
        """
        """
        # call the delivery method of the "current" city
        self.cities[-1].delivery(idents)
    def pickup(self, ident):
        """
        """
        # call the pickup method of the "current" city
        self.cities[-1].pickup(ident)

class YCHistoryCity:
    """
    """
    def __init__(self, name, layout, startingposition):
        """
        """
        # save the name
        self.name = name

        # save the layout
        self.layout = layout

        # save the starting position
        self.startingposition = startingposition

        # start with an empty list for deliveries
        self.deliveries = []
    def __len__(self):
        """
        """
        return len(self.deliveries)
    def __getitem__(self, index):
        """
        """
        # just index the deliveries list
        return self.deliveries[index]
    def delivery(self, idents):
        """
        """
        # create a delivery record with the provided ident tuples
        d = YCHistoryDelivery(idents)

        # append it to you deliveries list
        self.deliveries.append(d)

        # return it
        return d
    def pickup(self, ident):
        """
        """
        # call pickup on the "current" delivery
        self.deliveries[-1].pickup(ident)
    def getName(self):
        """
        """
        # return the name
        return self.name
    def getLayout(self):
        """
        """
        # return the layout
        return self.layout
    def getStartingPosition(self):
        """
        """
        # return the starting position
        return self.startingposition

class YCHistoryDelivery:
    """
    """
    def __init__(self, idents):
        """
        """
        # save the ident tuples
        self.idents = idents

        # the subject hasn't "chosen" a passenger yet
        self.pickedup = None
    def pickup(self, ident):
        """
        """
        # the subject has now "chosen" a passenger
        self.pickedup = ident
    def getIdentity(self):
        """
        """
        # return the passenger "chosen" by the subject
        return self.pickedup
    def getDestination(self):
        """
        """
        # search through the idents looking for the "chosen" passenger...
        for identity, location, destination in self.idents:
            # ...and when we find it...
            if identity == self.pickedup:
                # ...return its destination
                return destination

        # if we never found it, raise an exception
        raise LookupError, "Picked up passenger wasn't waiting"
    def getLocation(self):
        """
        """
        # search through the idents looking for the "chosen" passenger...
        for identity, location, destination in self.idents:
            # ...and when we find it...
            if identity == self.pickedup:
                # ...return its location
                return location

        # if we never found it, raise an exception
        raise LookupError, "Picked up passenger wasn't waiting"

class YCLayout:
    """
    """
    def __init__(self, xsize, ysize, config = None):
        """
        """
        if config is None:  # if no configuration, allow all spots to be either stores or buildings
            config = iter(lambda : "E", None)
        else:
            config = iter(config)
        self.rows = []
        for y in xrange(ysize):
            row = []
            self.rows.append(row)
            for x in xrange(xsize):
                row.append(config.next())
    def reconfigure(self, config):
        """
        """
        config = iter(config)
        self.rows = []
        for y in xrange(ysize):
            row = []
            self.rows.append(row)
            for x in xrange(xsize):
                row.append(config.next())
    def __eq__(self, other):
        """
        """
        for row1, row2 in izip(iter(self), iter(other)):
            if isinstance(row1, str) and isinstance(row2, str):
                if row1 != row2:
                    return False
                continue
            try:
                if row1.name != row2.name:
                    return False
            except AttributeError:
                return False
        return True
    def __getitem__(self, index):
        """
        """
        if not isinstance(index, tuple) or len(index) != 2:
            raise IndexError, "YCLayout indices must be two-dimensional coordinates"
        return self.rows[index[1]][index[0]]
    def __setitem__(self, index, value):
        """
        """
        if not isinstance(index, tuple) or len(index) != 2:
            raise IndexError, "YCLayout indices must be two-dimensional coordinates"
        self.rows[index[1]][index[0]] = value
    def __iter__(self):
        """
        """
        for row in self.rows:
            for value in row:
                yield value
    def getWidth(self):
        """
        """
        return len(self.rows[0])
    def getHeight(self):
        """
        """
        return len(self.rows)
    def iterGeometry(self):
        """
        """
        for x in self:
            if x == "S" or (hasattr(x, "type") and x.type == "store"):
                yield "S"
            elif x == "B" or (hasattr(x, "type") and x.type == "building"):
                yield "B"
            elif x == "N" or (hasattr(x, "type") and x.type == "nullbuilding"):
                yield "N"
            elif x == "NBNSW" or (hasattr(x, "type") and x.type == "nullbuildingnullsidewalk"):
                yield "NBNSW"
    def iterBuildingGeometry(self):
        """
        """
        for x in self:
            if x == "B" or (hasattr(x, "type") and x.type == "building"):
                yield x
            elif x == "N" or (hasattr(x, "type") and x.type == "nullbuilding"):
                yield "N"
            elif x == "NBNSW" or (hasattr(x, "type") and x.type == "nullbuildingnullsidewalk"):
                yield "NBNSW"
    def iterStoreGeometry(self):
        """
        """
        for x in self:
            if x == "S" or (hasattr(x, "type") and x.type == "store"):
                yield x
    def getGeometry(self):
        """
        """
        return list(self.iterGeometry())
    def getBuildingGeometry(self):
        """
        """
        return list(self.iterBuildingGeometry())
    def getStoreGeometry(self):
        """
        """
        return list(self.iterStoreGeometry())
    def iterStores(self):
        """
        """
        for x in self:
            if x == "S" or (hasattr(x, "type") and x.type == "store"):
                yield x
    def getStorePool(self):
        """
        """
        return Pool(*self.iterStores())
    def iterBuildings(self):
        """
        """
        for x in self:
            if x == "B" or (hasattr(x, "type") and x.type == "building"):
                yield x
    def getBuildingPool(self):
        """
        """
        return Pool(*self.iterBuildings())
    def fillBlanks(self, stores, buildings):
        """
        """
        # copy the pools...
        stores = stores[:]
        buildings = buildings[:]

        # eliminate stores and buildings already used...
        for thing in self:
            try:
                stores.remove(thing)
            except ValueError:
                pass
            try:
                buildings.remove(thing)
            except ValueError:
                pass

        storesleft = len(stores)
        buildingsleft = len(buildings)
        stores = iter(stores)
        buildings = iter(buildings)
        for y, row in enumerate(self.rows):
            for x, value in enumerate(row):
                if value == "S":
                    p = stores.next().copy()
                    p.type = "store"
                    self.rows[y][x] = p
                    storesleft -= 1
                elif value == "B":
                    p = buildings.next().copy()
                    p.type = "building"
                    self.rows[y][x] = p
                    buildingsleft -= 1
                elif value == "E":
                    r = random.randrange(storesleft + buildingsleft)
                    if r >= storesleft:
                        i = buildings
                        t = "building"
                        buildingsleft -= 1
                    else:
                        i = stores
                        t = "store"
                        storesleft -= 1
                    p = i.next().copy()
                    p.type = t
                    self.rows[y][x] = p
                elif value == "N":
                    p = PoolDict()
                    p.type = "nullbuilding"
                    self.rows[y][x] = p
                elif value == "NBNSW":
                    p = PoolDict()
                    p.type = "nullbuildingnullsidewalk"
                    self.rows[y][x] = p
    def addMeasurements(self, unitscale, wallheight, curbheight):
        """
        """
        cityrows = self.getHeight()
        citycolumns = self.getWidth()
        lowx = (cityrows - 1) * 3 * unitscale / 2
        lowz = (citycolumns - 1) * 3 * unitscale / 2
        self.realsize = ((cityrows * 2 * unitscale) + ((cityrows + 1) * unitscale),
                         (citycolumns * 2 * unitscale) + ((citycolumns + 1) * unitscale))
        self.unitscale = unitscale
        for rownum in xrange(cityrows):
            for columnnum in xrange(citycolumns):
                item = self[columnnum, rownum]
                xpos = ((unitscale * 3 * rownum) - lowx)
                zpos = (unitscale * 3 * columnnum) - lowz
                item.realpos = (xpos, 0.0, zpos)
                try:
                    item.realsize = (item.width * unitscale,
                                     item.height * unitscale,
                                     item.width * unitscale)
                except AttributeError:
                    pass
    def buildVE(self, vr, config, storetouchfunc):
        """
        """
        # start with a blank environment
        vr.resetEnvironment()
        
        # Sky...
        #vr.addSphere(0, 0, 0, config.skyImage, 1000, 3, 3)
        vr.addSkyBox(config.skyImage)

        # Floor & walls...
        vr.addFloorBox(0, 0, 0, self.realsize[0], config.wallHeight, self.realsize[1], config.roadImage, wallimage = config.wallImage, walltexlen = config.wallHeight / config.wallTexTimes, floortexlen = config.roadTexLen)
        
        # floor collisions...
        vr.setGravity(0.0, -0.1, 0.0)
        vr.addPlaneGeom(0.0, 1.0, 0.0, 0.0, mu = 0.0)

        # collisions for external walls...
        early_wall_distance=.01 #this is to prevent weird issues drawing the wall if we're essentially inside it - DO WE STILL NEED THIS?
        vr.addPlaneGeom(-1.0, 0.0, 0.0, -(early_wall_distance+self.realsize[0]/2.0), mu = 0)
        vr.addPlaneGeom(1.0, 0.0, 0.0, -(early_wall_distance+self.realsize[0]/2.0), mu = 0)
        vr.addPlaneGeom(0.0, 0.0, 1.0, -(early_wall_distance+self.realsize[1]/2.0), mu = 0)
        vr.addPlaneGeom(0.0, 0.0, -1.0, -(early_wall_distance+self.realsize[1]/2.0), mu = 0)

        # buildings and store collisions...
        for item in self:
            if item.type == "nullbuilding":
                xsize=config.unitScale * 2
                ysize=xsize + config.curbHeight
                zsize=config.unitScale * 2
                vr.addBuildingBox(item.realpos[0], 0.0, item.realpos[2], config.curbImage, xsize, config.curbHeight, roofimage = config.sidewalkImage, texlen = config.curbHeight, rooftexlen = config.sidewalkTexLen)
                vr.addBoxGeom(item.realpos[0], ysize / 2, item.realpos[2],
                              xsize + (2 * early_wall_distance), ysize, zsize + (2 * early_wall_distance), mu = 0)
            elif item.type == "nullbuildingnullsidewalk":
            #set these to zero so that the sidewalks disappear
                xsize=0.0
                ysize=0.0
                zsize=0.0
                vr.addBuildingBox(item.realpos[0], 0.0, item.realpos[2], config.curbImage, xsize, config.curbHeight, roofimage = config.sidewalkImage, texlen = config.curbHeight, rooftexlen = config.sidewalkTexLen)
                vr.addBoxGeom(item.realpos[0], ysize / 2, item.realpos[2],
                              xsize + (2 * early_wall_distance), ysize, zsize + (2 * early_wall_distance), mu = 0)
            elif item.type == "store":
                vr.addBuildingBox(item.realpos[0], 0.0, item.realpos[2], item.image, item.realsize[0], item.realsize[1])
                xsize=item.realsize[0]
                ysize=item.realsize[1]
                zsize=item.realsize[2]
                vr.addBoxGeom(item.realpos[0], ysize / 2, item.realpos[2], xsize + (2 * early_wall_distance), ysize,
                              zsize + (2 * early_wall_distance), mu = 0, callback = storetouchfunc, cbargs = (item.name,))
            else: #for buildings
                xsize=config.unitScale * 2
                ysize=item.realsize[1] + config.curbHeight
                zsize=config.unitScale * 2
                vr.addBuildingBox(item.realpos[0], config.curbHeight, item.realpos[2], item.image, item.realsize[0], item.realsize[1])
                vr.addBuildingBox(item.realpos[0], 0.0, item.realpos[2], config.curbImage, xsize, config.curbHeight, roofimage = config.sidewalkImage, texlen = config.curbHeight, rooftexlen = config.sidewalkTexLen)
                vr.addBoxGeom(item.realpos[0], ysize / 2, item.realpos[2],
                              xsize + (2 * early_wall_distance), ysize, zsize + (2 * early_wall_distance), mu = 0)
    def __copy__(self):
        """
        """
        return YCLayout(self.getWidth(), self.getHeight(), self)
    copy = __copy__  # normal copy method same as special one
