#!/usr/bin/python

from pyepl.locals import *

# other modules
import generate
import navigate
import sfp

# Set the current version
VERSION = '0.2.2'
MIN_PYEPL_VERSION = '1.0.6'

def prepare(exp, config):
    """
    """
    # get the state
    state = exp.restoreState()
    
    # start with an empty list for the sessions
    sessionList = []

    # start an empty pool for environments which transcends sessions
    bigEnvironmentPool = Pool()

    # make the history object
    bigHistory = generate.YCHistory()
    
    # for each session...
    for sessionNum in xrange(config.numSessions):
        # get session specific configuration
        sessionconfig = config.sequence(sessionNum)
        
        # start with an empty pool for the environments
        environmentPool = Pool()
        
        # for each environment...
        for environmentNum in xrange(sessionconfig.numEnvironments):
            # get trial specific configuration
            environmentconfig = sessionconfig.sequence(environmentNum)

            # get the environment name...
            try:
                envname = environmentconfig.environmentName
            except AttributeError:
                envname = "env%d_%d" % (sessionNum, environmentNum)
            
            # append environment name, stores list, passengers list, and buildings list to the environment pool
            bigEnvironmentPool.append(
                environmentPool.append(name = envname,
                                       stores = generate.chooseStores(environmentconfig, bigEnvironmentPool),
                                       passengers = generate.choosePassengers(environmentconfig, bigEnvironmentPool),
                                       buildings = generate.chooseBuildings(environmentconfig, bigEnvironmentPool),
                                       history = bigHistory
                                       )
                )

        # for each environment...
        for environmentNum in xrange(sessionconfig.numEnvironments):
            # get trial specific configuration
            environmentconfig = sessionconfig.sequence(environmentNum)
            
            # prepare store front practice for this environment...
            if environmentconfig.SFPMode is None:
                # this environment has no store front practice, so make its state None
                environmentPool[environmentNum].SFPState = None
                
            elif environmentconfig.SFPMode == "timedVolbit":
                # prepare a timed volbit session
                environmentPool[environmentNum].SFPState = sfp.prepareVolbit(environmentconfig, environmentPool)
                
            elif environmentconfig.SFPMode == "untimedVolbit":
                # prepare a untimed volbit session
                environmentPool[environmentNum].SFPState = sfp.prepareVolbit(environmentconfig, environmentPool)
                
            else:
                # since it didn't match any of our mode names, raise an exception
                raise NotImplementedError, "Unrecognized store front practice mode: %s" % environmentconfig.SFPMode
            
        # append the environments to the session list
        sessionList.append(environmentPool)

    # save the information
    exp.saveState(state,
                  environmentNum = 0,
                  SFPDone = False,
                  environmentState = State(),
                  sessionList = sessionList,
                  sessionNum = 0)

def run(exp, config):
    """
    """
    # get the state
    state = exp.restoreState()
    
    # set up session...
    if state.sessionNum >= len(state.sessionList):
        # if all the sessions have been run, don't continue
        print "No more sessions!"
        return

    # set the session number (so PyEPL knows what directory to put the data in)
    exp.setSession(state.sessionNum)

    # get session specific configuration
    sessionconfig = config.sequence(state.sessionNum)

    # create tracks...
    video = VideoTrack("video")
    vr = VRTrack("vr")
#    audio = AudioTrack("audio")
    keyboard = KeyTrack("keyboard")
#    mouse = MouseTrack("mouse")
    joystick = JoyTrack("joystick")

    # calibrate the joystick
    joystick.calibrate()
    
    # create a PresentationClock to handle timing
    clock = PresentationClock()

    # do each environment...
    while state.environmentNum < len(state.sessionList[state.sessionNum]):
        # get trial specific configuration
        environmentconfig = sessionconfig.sequence(state.environmentNum)

        # unpack the components of the environment
        env = state.sessionList[state.sessionNum][state.environmentNum]

        #break instructions...
        video.clear("black")
        if environmentconfig.breakInstructions:
            instruct(environmentconfig.breakInstructions,
                     font = environmentconfig.instructionsFont,
                     size = environmentconfig.instructionsFontSize,
                     color = environmentconfig.instructionsFontColor,
                     justification = environmentconfig.instructionsJustification,
                     topmargin = environmentconfig.instructionsTopMargin,
                     leftmargin = environmentconfig.instructionsLeftMargin,
                     rightmargin = environmentconfig.instructionsRightMargin,
                     clk = clock)

        # if we haven't done the store front practice for this environment...
        if not state.SFPDone:
            # ...do it now...

            # SFP instructions...
            video.clear("black")
            if environmentconfig.SFPInstructions:
                instruct(environmentconfig.SFPInstructions,
                         font = environmentconfig.instructionsFont,
                         size = environmentconfig.instructionsFontSize,
                         color = environmentconfig.instructionsFontColor,
                         justification = environmentconfig.instructionsJustification,
                         topmargin = environmentconfig.instructionsTopMargin,
                         leftmargin = environmentconfig.instructionsLeftMargin,
                         rightmargin = environmentconfig.instructionsRightMargin,
                         clk = clock)
            
            if environmentconfig.SFPMode is None:
                # this environment has no store front practice, so just continue on the the city
                pass
                
            elif environmentconfig.SFPMode == "timedVolbit":
                # run the timed volbit session
                sfp.runTimedVolbit(environmentconfig, env.SFPState, clock, video)
                
            elif environmentconfig.SFPMode == "untimedVolbit":
                # run the untimed volbit session
                sfp.runUntimedVolbit(environmentconfig, env.SFPState, clock, video)
                
            else:
                # since it didn't match any of our mode names, raise an exception
                raise NotImplementedError, "Unrecognized store front practice mode: %s" % environmentconfig.SFPMode

            # now that we've done the store front practice, note that in the state
            state = exp.saveState(state, SFPDone = True)

        # navigation instructions...
        video.clear("black")
        if environmentconfig.navigationInstructions:
            instruct(environmentconfig.navigationInstructions,
                     font = environmentconfig.instructionsFont,
                     size = environmentconfig.instructionsFontSize,
                     color = environmentconfig.instructionsFontColor,
                     justification = environmentconfig.instructionsJustification,
                     topmargin = environmentconfig.instructionsTopMargin,
                     leftmargin = environmentconfig.instructionsLeftMargin,
                     rightmargin = environmentconfig.instructionsRightMargin,
                     clk = clock)

        # run the environment navigation
        navigate.runCity(environmentconfig, env, clock, state.environmentState)

        # ...and save the state so that the experiment can be restored from this point
        state = exp.saveState(state, environmentNum = state.environmentNum + 1, SFPDone = False)

    # end screen...
    video.clear("black")
    if environmentconfig.youAreDoneMessage:
        instruct(environmentconfig.youAreDoneMessage,
                 font = environmentconfig.instructionsFont,
                 size = environmentconfig.instructionsFontSize,
                 color = environmentconfig.instructionsFontColor,
                 justification = environmentconfig.instructionsJustification,
                 topmargin = environmentconfig.instructionsTopMargin,
                 leftmargin = environmentconfig.instructionsLeftMargin,
                 rightmargin = environmentconfig.instructionsRightMargin,
                 clk = clock)

    # save the state when the session is finished
    # with trial and session numbers set for the next session
    exp.saveState(state, environmentNum = 0, sessionNum = state.sessionNum + 1)

    # wait for the clock to catch up
    clock.wait()


# only do this if the experiment is run as a stand-alone program (not imported as a library)...
if __name__ == "__main__":
    # make sure we have the min PyEPL version
    checkVersion(MIN_PYEPL_VERSION)
    # set up the experiment and disable screen syncing...
    exp = Experiment(sync_to_vbl=False)
    exp.parseArgs()
    exp.setup()

    # allow users to break out of the experiment with escape-F1 (the default key combo)
    exp.setBreak()

    # get the subject configuration
    config = exp.getConfig()

    # if there was no saved state, run the prepare function
    if not exp.restoreState():
        prepare(exp, config)

    # now run the subject
    run(exp, config)
